(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.DataUtils = {}));
}(this, (function (exports) { 'use strict';

	/**
	 * adds two numbers
	 * @example
	 * add(1, 2)
	 * // returns 3
	 */
	function add(first, second) {
	  return first + second;
	}

	var _ref = {},
	  hasOwnProperty = _ref.hasOwnProperty;
	/**
	 * returns true or false depending if the provided property is present inside the provided object
	 * @example
	 * hasOwn('a', { a: 1, b: 2 })
	 * // returns true
	 */
	function hasOwn(prop, obj) {
	  return hasOwnProperty.call(obj, prop);
	}

	/**
	 * assigns properties of one or more source objects onto target object
	 * @example
	 * assign({ a: 1, b: 2 }, { b: 4, c: 5 })
	 * // returns { a: 1, b: 4, c: 5 }
	 */
	function assign() {
	  // eslint-disable-next-line @typescript-eslint/ban-ts-comment
	  // @ts-ignore
	  // eslint-disable-next-line no-func-assign
	  assign = Object.assign || function (target) {
	    for (var _len = arguments.length, sources = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
	      sources[_key - 1] = arguments[_key];
	    }
	    sources.forEach(function (source) {
	      for (var key in source) {
	        if (hasOwn(key, source)) {
	          target[key] = source[key];
	        }
	      }
	    });
	    return target;
	  };
	  return assign.apply(void 0, arguments);
	}

	/**
	 * capitalizes first letter of string
	 * @example
	 * capitalizeFirstLetter('hello')
	 * // returns 'Hello'
	 */
	function capitalizeFirstLetter(text) {
	  return text.charAt(0).toUpperCase() + text.slice(1);
	}

	/**
	 * returns a new array formed by applying a given callback function to each element of the array, and then flattening the result by one level
	 * @example
	 * const arr = [{ a: [1, 2] }, { a: [3, 4] }, { a: [5, 6] }]
	 * flatMap(el => el.a, arr)
	 * // returns [1, 2, 3, 4, 5, 6]
	 */
	function flatMap(iteratee, arr) {
	  var _ref;
	  return (_ref = []).concat.apply(_ref, arr.map(iteratee));
	}

	/**
	 * takes two arguments: a function and either an array or another function. The function applies the input function to every element of the input array or to the result of calling the input function on the input array. Then, the results are concatenated into a new array and returned.
	 * @example
	 * chain(n => [n, n], [1, 2, 3])
	 * // returns [1, 1, 2, 2, 3, 3]
	 * const head = list => list[0]
	 * const append = el => list => list.concat(el)
	 * chain(append, head)([1, 2, 3])
	 * // returns [1, 2, 3, 1]
	 */
	function chain(fn, listOrFn) {
	  if (typeof listOrFn === 'function') {
	    return function (list) {
	      return fn(listOrFn(list))(list);
	    };
	  }
	  return flatMap(fn, listOrFn);
	}

	/**
	 * chunks provided array to specified size chunks
	 * @example
	 * chunk([1, 2, 3, 4], 2)
	 * // returns [[1, 2], [3, 4]]
	 */
	function chunk(arr, number) {
	  if (number <= 0) {
	    return [];
	  }
	  var chunked = [];
	  for (var i = 0; i < arr.length; i += number) {
	    chunked.push(arr.slice(i, i + number));
	  }
	  return chunked;
	}

	/**
	 * determines whether provided value is an array
	 * @example
	 * isArray([1, 2])
	 * // returns true
	 * isArray('hello')
	 * // returns false
	 */
	function isArray(arr) {
	  return Array.isArray(arr);
	}

	/**
	 * returns true or false depending if the provided value is an object
	 * @example
	 * isObject({ a: 1 })
	 * // returns true
	 * isObject([1, 2])
	 * // returns false
	 */
	// eslint-disable-next-line @typescript-eslint/ban-types
	function isObject(obj) {
	  return typeof obj === 'object' && obj !== null && !isArray(obj);
	}

	/**
	 * returns an array of the provided object keys
	 * @example
	 * keys({ a: 1, b: 2, c: 3 })
	 * // returns ['a', 'b', 'c']
	 */
	function keys(obj) {
	  if ('keys' in Object && typeof Object.keys === 'function') {
	    return Object.keys(obj);
	  }
	  var keysArray = [];
	  for (var property in obj) {
	    if (Object.prototype.hasOwnProperty.call(obj, property)) {
	      keysArray.push(property);
	    }
	  }
	  return keysArray;
	}

	/**
	 * maps values of the provided object
	 * @example
	 * mapValues(val => val.toUpperCase(), { a: 'foo', b: 'bar' })
	 * // returns { a: 'FOO', b: 'BAR' }
	 */
	function mapValues(mapper, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    acc[key] = mapper(obj[key]);
	    return acc;
	  }, {});
	}

	/**
	 * returns an exact clone of the provided value with a new reference
	 * @example
	 * cloneDeep({ a: 1, b: 2 })
	 * // returns { a: 1, b: 2 }
	 */
	function cloneDeep(value) {
	  if (isArray(value)) {
	    return value.map(cloneDeep);
	  }
	  if (isObject(value)) {
	    return mapValues(cloneDeep, value);
	  }
	  return value;
	}

	// slightly simplified version of https://github.com/sindresorhus/camelcase/blob/a526ef0399f9a1310eaacafa0ae4a69da4a2f1ad/index.js
	// TODO: possibly could be rewritten using short-ish regexp

	function preserveCamelCase(input) {
	  var text = input;
	  var isLastCharLower = false;
	  var isLastCharUpper = false;
	  var isLastLastCharUpper = false;
	  for (var index = 0; index < text.length; index++) {
	    var _char = text[index];
	    if (isLastCharLower && /[a-zA-Z]/.test(_char) && _char.toUpperCase() === _char) {
	      text = text.slice(0, index) + '-' + text.slice(index);
	      isLastCharLower = false;
	      isLastLastCharUpper = isLastCharUpper;
	      isLastCharUpper = true;
	      index++;
	    } else if (isLastCharUpper && isLastLastCharUpper && /[a-zA-Z]/.test(_char) && _char.toLowerCase() === _char) {
	      text = text.slice(0, index - 1) + '-' + text.slice(index - 1);
	      isLastLastCharUpper = isLastCharUpper;
	      isLastCharUpper = false;
	      isLastCharLower = true;
	    } else {
	      isLastCharLower = _char.toLowerCase() === _char;
	      isLastLastCharUpper = isLastCharUpper;
	      isLastCharUpper = _char.toUpperCase() === _char;
	    }
	  }
	  return text;
	}

	/**
	 * converts provided string to camel case
	 * @example
	 * camelCase('hello_world')
	 * // returns 'helloWorld'
	 */
	function camelCase(input) {
	  var text = input.trim();
	  if (text.length === 0) {
	    return '';
	  }
	  if (text.length === 1) {
	    return text.toLowerCase();
	  }
	  if (/^[a-z\d]+$/.test(text)) {
	    return text;
	  }
	  var hasUpperCase = text !== text.toLowerCase();
	  if (hasUpperCase) {
	    text = preserveCamelCase(text);
	  }
	  text = text.replace(/^[_.\- ]+/, '').toLowerCase().replace(/[_.\- ]+(\w|$)/g, function (match, p1) {
	    return p1.toUpperCase();
	  });
	  return text;
	}

	/**
	 * removes all nullish and NaN values from the provided object or array
	 * @example
	 * compact({ a: 1, b: null })
	 * // returns { a: 1 }
	 * compact(['hello', undefined])
	 * // returns ['hello']
	 */
	function compact(collection) {
	  return isArray(collection) ? collection.filter(function (value) {
	    return value !== null && value !== undefined && !Number.isNaN(value);
	  }) : Object.keys(collection).reduce(function (result, key) {
	    var value = collection[key];
	    if (value !== null && value !== undefined && !Number.isNaN(value)) {
	      result[key] = value;
	    }
	    return result;
	  }, {});
	}

	/**
	 * returns a composed value from the provided arguments
	 * @example
	 * const f3 = (a: string) => `f3(${a})`
	 * const f2 = (a: string) => `f2(${a})`
	 * const f1 = (a: string) => `f1(${a})`
	 * compose(f3, f2, f1)('arg')
	 * // returns 'f3(f2(f1(arg)))'
	 */
	function compose() {
	  for (var _len = arguments.length, funcs = new Array(_len), _key = 0; _key < _len; _key++) {
	    funcs[_key] = arguments[_key];
	  }
	  return funcs.reduce(function (composed, next) {
	    return function () {
	      return composed(next.apply(void 0, arguments));
	    };
	  });
	}

	/**
	 * delays the execution of a function until a certain period of time has passed without the function being called again
	 * @example
	 * debounce(1000, someFunction)
	 */
	function debounce(ms, fn) {
	  // actual return type of setTimeout differs per platform (browser vs node)
	  var timeoutId;
	  return function () {
	    clearTimeout(timeoutId);
	    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
	      args[_key] = arguments[_key];
	    }
	    timeoutId = setTimeout.apply(void 0, [fn, ms].concat(args));
	  };
	}

	/**
	 * recurses through an object and transforms its keys – and the keys of any nested objects – based on the provided key mapping function
	 * @example
	 * deepMapKeys(key => `test_${key}`, { a: 1, b: { c: 2 } })
	 * // returns { test_a: 1, test_b: { test_c: 2 } }
	 */
	function deepMapKeys(mapper, obj) {
	  return Object.keys(obj).reduce(function (acc, key) {
	    if (obj[key] && typeof obj[key] === 'object') {
	      if (Array.isArray(obj[key])) {
	        acc[mapper(key)] = obj[key].map(function (el) {
	          return deepMapKeys(mapper, el);
	        });
	        return acc;
	      }
	      acc[mapper(key)] = deepMapKeys(mapper, obj[key]);
	      return acc;
	    }
	    acc[mapper(key)] = obj[key];
	    return acc;
	  }, {});
	}

	/**
	 * deep merges provided objects by assigning the sources onto the target object passed as the first argument
	 * @example
	 * deepMerge({ a: 1, b: { c: 2 } }, { b: { d: 3 } })
	 * // returns { a: 1, b: { c: 2, d: 3 } }
	 */
	function deepMerge(target) {
	  for (var _len = arguments.length, sources = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
	    sources[_key - 1] = arguments[_key];
	  }
	  if (!sources.length) return target;
	  var source = sources.shift();
	  if (isObject(target) && isObject(source)) {
	    for (var key in source) {
	      if (isObject(source[key])) {
	        var _Object$assign;
	        if (!target[key]) Object.assign(target, (_Object$assign = {}, _Object$assign[key] = {}, _Object$assign));
	        deepMerge(target[key], source[key]);
	      } else {
	        var _Object$assign2;
	        Object.assign(target, (_Object$assign2 = {}, _Object$assign2[key] = source[key], _Object$assign2));
	      }
	    }
	  }
	  return deepMerge.apply(void 0, [target].concat(sources));
	}

	/**
	 * returns true or false depending if the provided value is null or undefined
	 * @example
	 * isNil(null)
	 * // returns true
	 */
	function isNil(value) {
	  return value === null || value === undefined;
	}

	/**
	 * returns a function that invoked with a null or undefined values returns the provided default
	 * @example
	 * const defaulter = defaultTo('default')
	 * defaulter('hello')
	 * // returns 'hello'
	 * defaulter(null)
	 * // returns 'default'
	 */
	function defaultTo(defaultValue) {
	  return function (value) {
	    return isNil(value) ? defaultValue : value;
	  };
	}

	/**
	 * returns an array of keys of the object passed as the first argument that are not present, or that are present, but have different values, in object passed as the second argument
	 * @example
	 * diffKeys({ a: 1, b: 2, c: 3 }, { c: 3, d: 4 })
	 * // returns ['a', 'b']
	 */
	function diffKeys(obj, comparable) {
	  return keys(obj).reduce(function (diffKeys, key) {
	    if (obj[key] !== comparable[key]) {
	      diffKeys.push(key);
	    }
	    return diffKeys;
	  }, []);
	}

	/**
	 * drops specific amount of elements of the provided array starting at the beginning
	 * @example
	 * drop(2, [1, 2, 3, 4])
	 * // returns [3, 4]
	 */
	function drop(count, arr) {
	  return arr.slice(count);
	}

	/**
	 * drops specific amount of elements of the provided array starting at the end
	 * @example
	 * dropRight(2, [1, 2, 3, 4])
	 * // returns [1, 2]
	 */
	function dropRight(count, arr) {
	  return arr.slice(0, -count);
	}

	/**
	 * returns an array if an array was provided, or a new array if provided value was not an array
	 * @example
	 * ensureArray([1, 2])
	 * // returns [1, 2]
	 * ensureArray('test')
	 * // returns ['test']
	 */
	function ensureArray(maybeArr) {
	  return isArray(maybeArr) ? maybeArr : [maybeArr];
	}

	/**
	 * returns an array of entries of the provided object
	 * @example
	 * entries({ a: 1, b: 2 })
	 * // returns [['a', 1], ['b', 2]]
	 */
	function entries(obj) {
	  return keys(obj).map(function (key) {
	    return [key, obj[key]];
	  });
	}

	/**
	 * returns the first element of the provided array that returns true for the predicate function, or undefined if the element was not found
	 * @example
	 * find(el => el === 2, [1, 2, 3])
	 * // returns 2
	 */
	function find(predicate, arr) {
	  for (var index = 0; index < arr.length; index++) {
	    var element = arr[index];
	    if (predicate(element)) {
	      return element;
	    }
	  }
	}

	/**
	 * returns the index of the element of the provided array that returns true for the predicate function, or -1 if the element was not found
	 * @example
	 * findIndex(el => el === 2, [1, 2, 3])
	 * // returns 1
	 * findIndex(el => el === 5, [1, 2, 3])
	 * // returns -1
	 */
	function findIndex(predicate, arr) {
	  for (var index = 0; index < arr.length; index++) {
	    if (predicate(arr[index])) {
	      return index;
	    }
	  }
	  return -1;
	}

	/**
	 * returns the key of the provided object that returns true for the predicate function, or undefined if the key was not found
	 * @example
	 * findKey(el => el === 1, { a: 1, b: { c: 2 } })
	 * // returns 'a'
	 * findKey(el => el === 2, { a: 1, b: { c: 2 } })
	 * // returns undefined
	 */
	function findKey(predicate, obj) {
	  return find(function (key) {
	    return predicate(obj[key]);
	  }, keys(obj));
	}

	/**
	 * returns the last element of the provided array that returns true for the predicate function, or undefined if the element was not found
	 * @example
	 * findLast(el => el % 2 === 0, [1, 2, 3, 4])
	 * // returns 4
	 */
	function findLast(predicate, arr) {
	  for (var index = arr.length - 1; index >= 0; index--) {
	    if (predicate(arr[index])) {
	      return arr[index];
	    }
	  }
	}

	/**
	 * returns the index of the last element of the provided array that returns true for the predicate function, starting from the specified index element to the begining of the array, or -1 if the element was not found
	 * @example
	 * findLastIndexFrom(el => el % 2 === 0, 3, [1, 2, 3, 4, 5, 6])
	 * // returns 3
	 */
	function findLastIndexFrom(predicate, startIndex, arr) {
	  for (var index = startIndex; index >= 0; index--) {
	    if (predicate(arr[index])) {
	      return index;
	    }
	  }
	  return -1;
	}

	/**
	 * returns the index of the last element of the provided array that returns true for the predicate function, or -1 if the element was not found
	 * @example
	 * findLastIndex(el => el % 2 === 0, [1, 2, 3, 4])
	 * // returns 3
	 */
	function findLastIndex(predicate, arr) {
	  return findLastIndexFrom(predicate, arr.length - 1, arr);
	}

	/**
	 * returns an array of requested length filled with provided value
	 * @example
	 * filledArray(3, 1)
	 * // returns [1, 1, 1]
	 */
	function filledArray(length, value) {
	  if (length <= 0) {
	    return [];
	  }
	  var arr = [];
	  while (length--) {
	    arr.push(value);
	  }
	  return arr;
	}

	/**
	 * returns the provided value
	 * @example
	 * identity('hello')
	 * // returns 'hello'
	 */
	function identity(value) {
	  return value;
	}

	/**
	 * flattens the provided array by one level
	 * @example
	 * flatten([1, 2, [3, 4]])
	 * // returns [1, 2, 3, 4]
	 * flatten([1, 2, [3, [4, 5]], 6])
	 * // returns [1, 2, 3, [4, 5], 6]
	 */
	function flatten(arr) {
	  return flatMap(identity, arr);
	}

	/**
	 * invokes the provided callback for each of the provided object fields
	 * @example
	 * const obj = { a: 1, b: 2, c: 3 }
	 * const arr: string[] = []
	 * forOwn(el => arr.push(el.toString()), obj)
	 * // arr equals ['1', '2', '3']
	 */
	function forOwn(callback, obj) {
	  return keys(obj).forEach(function (key) {
	    callback(obj[key], key);
	  });
	}

	/**
	 * returns an object constructed from the provided array of key, value pairs'
	 * @example
	 * const arr = [['a', 1], ['b', 2], ['c', 3]]
	 * fromPairs(arr)
	 * // returns { a: 1, b: 2, c: 3 }
	 */
	function fromPairs(pairs) {
	  return pairs.reduce(function (obj, _ref) {
	    var key = _ref[0],
	      value = _ref[1];
	    obj[key] = value;
	    return obj;
	  }, {});
	}

	/**
	 * generates a random id
	 * @example
	 * generateRandomId()
	 * // returns 'd1rjknhhch8'
	 */
	function generateRandomId() {
	  return Math.random().toString(36).substring(2);
	}

	/**
	 * generates an unique id based on the provided object keys
	 * @example
	 * generateUniqueId({ xuvarw8cao: 1, b837g2nba1d: 2 })
	 * // returns 'd1rjknhhch8'
	 */
	function generateUniqueId(map) {
	  var id = generateRandomId();
	  return hasOwn(id, map) ? generateUniqueId(map) : id;
	}

	/**
	 * returns the value from the specififed path from the provided object
	 * @example
	 * const obj = { a: { b: [1, 2, 3] } }
	 * get('a.b.1', obj)
	 * // returns 2
	 * get(['a', 'b', '1'], obj)
	 * // returns 2
	 */
	function get(propPath, obj) {
	  var arrPath = typeof propPath === 'string' ? propPath.split('.') : propPath;
	  var pathPartIndex = 0;
	  var result = obj;
	  while (result && pathPartIndex < arrPath.length) {
	    result = result[arrPath[pathPartIndex++]];
	  }
	  return result;
	}

	/**
	 * returns the value from the specififed path from the provided object or the default value if the path element was not found
	 * @example
	 * const obj = { a: { b: [1, 2, 3] } }
	 * getOr(null, 'a.b.1', obj)
	 * // returns 2
	 * getOr(null, 'a.c', obj)
	 * // returns null
	 */
	function getOr(defaultValue, prop, obj) {
	  var propValue = get(prop, obj);
	  return propValue !== undefined && propValue !== null ? propValue : defaultValue;
	}

	/**
	 * groups values from the provided array or object based on the result of the provided mapper function
	 * @example
	 * const arr = [
	 * 	{ a: 1, b: 2 },
	 * 	{ a: 2, b: 4 },
	 * 	{ a: 3, b: 6 },
	 * ]
	 * groupBy(el => (el.a % 2 === 0 ? 'even' : 'odd'), arr)
	 * // returns {
	 * // 	odd: [
	 * // 		{ a: 1, b: 2 },
	 * // 		{ a: 3, b: 6 },
	 * // 	],
	 * // 	even: [{ a: 2, b: 4 }],
	 * // }
	 */
	function groupBy(mapper, collection) {
	  return Object.keys(collection).reduce(function (grouped, key) {
	    var element = collection[key];
	    var groupKey = mapper(element);
	    grouped[groupKey] = grouped[groupKey] || [];
	    grouped[groupKey].push(element);
	    return grouped;
	  }, {});
	}

	/**
	 * groups values from the provided object based on the result of the provided key mapping function
	 * @example
	 * const obj = { a: 1, b: 2, c: 3 }
	 * groupKeys(el => (el === 'a' ? 'aaa' : 'rest'), obj)
	 * // returns {
	 * //	aaa: { a: 1 },
	 * //	rest: { b: 2, c: 3 },
	 * // }
	 */
	function groupKeys(mapper, obj) {
	  return keys(obj).reduce(function (grouped, key) {
	    var groupKey = mapper(key);
	    grouped[groupKey] = grouped[groupKey] || {};
	    grouped[groupKey][key] = obj[key];
	    return grouped;
	  }, {});
	}

	/**
	 * returns true or false depending if the provided value is present inside the provided array or string
	 * @example
	 * includes('a', ['a', 'b', 'c'])
	 * // returns true
	 * includes('d', 'abc')
	 * // returns false
	 */
	function includes(value, arrOrStr) {
	  return arrOrStr.indexOf(value) !== -1;
	}

	/**
	 * returns true or false depending if the provided value is an empty object or an empty array
	 * @example
	 * isEmpty({})
	 * // returns true
	 */
	function isEmpty(collection) {
	  return (isArray(collection) ? collection : Object.keys(collection)).length === 0;
	}

	/**
	 * returns true or false depending if the provided value is falsy
	 * @example
	 * isFalsy(0)
	 * // returns true
	 */
	function isFalsy(value) {
	  return !value;
	}

	/**
	 * returns true or false depending if the provided value is truthy
	 * @example
	 * isTruthy(1)
	 * // returns true
	 */
	function isTruthy(value) {
	  return !!value;
	}

	/**
	 * returns true or false depending if the provided value is a Promise
	 * @example
	 * isPromise(new Promise(res => res(null)))
	 * // returns true
	 */
	function isPromise(promise) {
	  return !!promise && typeof promise.then === 'function';
	}

	/**
	 * constructs an object from the provided array of objects, grouped by the value of the specified key
	 * @example
	 * const arr = [
	 * 	{ a: 'foo', b: 'bar' },
	 * 	{ a: 'foo', b: 'baz' },
	 * 	{ a: 'test', b: 'bab' },
	 * ]
	 * keyBy('a', arr)
	 * // returns { foo: { a: 'foo', b: 'baz' }, test: { a: 'test', b: 'bab' } }
	 */
	function keyBy(prop, arr) {
	  return arr.reduce(function (acc, el) {
	    acc[el[prop]] = el;
	    return acc;
	  }, {});
	}

	/**
	 * returns the last element of the provided array or undefined when array is empty
	 * @example
	 * last([1, 2, 3])
	 * // returns 3
	 */
	function last(arr) {
	  return arr.length > 0 ? arr[arr.length - 1] : undefined;
	}

	/**
	 * maps keys of the provided object
	 * @example
	 * mapKeys(key => key.toUpperCase(), { a: 1, b: 2 })
	 * // returns { A: 1, B: 2 }
	 */
	function mapKeys(mapper, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    Object.defineProperty(acc, mapper(key), {
	      value: obj[key],
	      enumerable: true
	    });
	    return acc;
	  }, {});
	}

	/**
	 * maps values and keys of the provided object
	 * @example
	 * mapValuesIndexed((val, key) => `${key}-${val}`, { a: 1, b: 2 })
	 * // returns { a: 'a-1', b: 'b-2' }
	 */
	function mapValuesIndexed(iteratee, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    acc[key] = iteratee(obj[key], key);
	    return acc;
	  }, {});
	}

	/**
	 * deep merges the two provided objects
	 * @example
	 * merge({ a: 1 }, { b: 2 })
	 * // returns { a: 1, b: 2 }
	 */
	function merge(objA, objB) {
	  if (isEmpty(objB)) {
	    return objA;
	  }
	  var result = {};
	  forOwn(function (value, key) {
	    if (hasOwn(key, objB)) {
	      if (isObject(objA[key]) && isObject(objB[key])) {
	        result[key] = merge(objA[key], objB[key]);
	      } else if (isArray(objA[key]) && isArray(objB[key])) {
	        var length = Math.max(objA[key].length, objB[key].length);
	        result[key] = new Array(length);
	        for (var i = 0; i < length; i++) {
	          if (i in objB[key]) {
	            result[key][i] = objB[key][i];
	          } else if (i in objA[key]) {
	            result[key][i] = objA[key][i];
	          }
	        }
	      } else {
	        result[key] = objB[key];
	      }
	    } else {
	      result[key] = objA[key];
	    }
	  }, objA);
	  forOwn(function (value, key) {
	    if (!hasOwn(key, result)) {
	      result[key] = objB[key];
	    }
	  }, objB);
	  return result;
	}

	/**
	 * deep merges all provided objects
	 * @example
	 * mergeAll({ a: 1 }, { b: 2 }, { c: 3 })
	 * // returns { a: 1, b: 2, c: 3 }
	 */
	function mergeAll(objs) {
	  if (objs.length === 0) {
	    return {};
	  }
	  var first = objs[0],
	    rest = objs.slice(1);
	  return rest.reduce(function (merged, obj) {
	    return merge(merged, obj);
	  }, first);
	}

	/**
	 * memoizes the provided function so it returns cached results but based on the provided key resolver
	 * @example
	 * const memoizedFunc = memoizeWith(key => (key === 'a' ? key : 'bar'), (val: string) => ({ current: val }))
	 * memoizedFunc('a') === memoizedFunc('a') // true
	 * memoizedFunc('b') === memoizedFunc('c') // true
	 */
	function memoizeWith(keyResolver, func) {
	  var cache = {};
	  return function () {
	    var key = keyResolver.apply(void 0, arguments);
	    if (hasOwn(key, cache)) {
	      return cache[key];
	    }
	    var value = func.apply(void 0, arguments);
	    cache[key] = value;
	    return value;
	  };
	}

	// TODO: technically this should accept AnyFunction but it doesn't type check currently with that for some reason

	/**
	 * memoizes the provided function so it returns cached results when invoked with the same arguments
	 * @example
	 * const memoizedFunc = memoize((val: number) => ({ current: val })
	 * memoizedFunc(3) === memoizedFunc(3)
	 * // returns true
	 */
	function memoize(func) {
	  return memoizeWith(identity, func);
	}

	/**
	 * memoizes the provided function so it returns cached results but only with one cache key
	 * @example
	 * const memoizedFunc = memoizeOne((val: number) => ({ current: val })
	 *
	 * const resultFor1 = memoizedFunc(1) // { current: 1 }
	 * resultFor1 === memoizedFunc(1) // true
	 * resultFor1 === memoizedFunc(1) // true
	 *
	 * const resultFor2 = memoizedFunc(2) // { current: 2 }
	 * resultFor2 === memoizedFunc(2) // true
	 * resultFor1 === memoizedFunc(1) // false
	 */
	function memoizeOne(fn) {
	  var called = false;
	  var memoized;
	  var cacheKey;
	  return function () {
	    if (called && (arguments.length <= 0 ? undefined : arguments[0]) === cacheKey) {
	      return memoized;
	    }
	    called = true;
	    cacheKey = arguments.length <= 0 ? undefined : arguments[0];
	    memoized = fn.apply(void 0, arguments);
	    return memoized;
	  };
	}

	/**
	 * does literally nothing
	 * @example
	 * somethingAsyncAndDangerous().catch(noop)
	 */
	// eslint-disable-next-line @typescript-eslint/no-empty-function
	function noop() {}

	/**
	 * returns an array of values of the provided object
	 * @example
	 * values({ a: 1, b: 2, c: 3 })
	 * // returns [1, 2, 3]
	 */
	function values(obj) {
	  return keys(obj).map(function (key) {
	    return obj[key];
	  });
	}

	/**
	 * sorts values of the provided object or array based on the provided mapping function or the provided prop string
	 * @example
	 * const obj = {
	 * 	a: { chats: 1 },
	 * 	b: { chats: 3 },
	 * 	c: { chats: 2 },
	 * }
	 * numericSortBy(el => el.chats * -1, obj)
	 * // returns [{ chats: 3 }, { chats: 2 }, { chats: 1 }]
	 */
	function numericSortBy(propOrMapper, collection) {
	  var mapper = typeof propOrMapper === 'function' ? propOrMapper : function (element) {
	    return get(propOrMapper, element);
	  };
	  return (isArray(collection) ? [].concat(collection) : values(collection)).sort(function (elementA, elementB) {
	    return mapper(elementA) - mapper(elementB);
	  });
	}

	/**
	 * returns an object the same as the provided one but without the fields omitted by the predicate function
	 * @example
	 * omitByIndexed((val, key) => key === 'b', { a: 1, b: 2, c: 3 })
	 * // returns { a: 1, c: 3 }
	 */
	function omitByIndexed(predicate, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    if (!predicate(obj[key], key)) {
	      acc[key] = obj[key];
	    }
	    return acc;
	  }, {});
	}

	/**
	 * returns the provided object but with specified keys omitted
	 * @example
	 * omit(['a'], { a: 1, b: 2 })
	 * // returns { b: 2 }
	 */
	function omit(keys, obj) {
	  return omitByIndexed(function (_, key) {
	    return keys.indexOf(key) !== -1;
	  }, obj);
	}

	/**
	 * returns an object same as the provided one but without the fields omitted by the predicate function
	 * @example
	 * omitBy(el => el % 2 === 0, { a: 1, b: 2, c: 3 })
	 * // returns { a: 1, c: 3 }
	 */
	function omitBy(predicate, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    if (!predicate(obj[key])) {
	      acc[key] = obj[key];
	    }
	    return acc;
	  }, {});
	}

	/**
	 * returns a particular function to be executed only a single time
	 * @example
	 * const results = [1, 2, 3, 4, 5]
	 * const getResult = once(() => results.pop())
	 * getResult() // 5
	 * getResult() // 5
	 * getResult() // 5
	 *
	 */
	function once(fn) {
	  var called = false;
	  var result;
	  return function () {
	    if (called) {
	      return result;
	    }
	    called = true;
	    return result = fn.apply(void 0, arguments);
	  };
	}

	/**
	 * returns a function that called invokes the provided array of functions with the call arguments
	 * @example
	 * const func = over([val => val + 1, val => val * 10, Boolean])
	 * func(1)
	 * // returns [2, 10, true]
	 */
	function over(fns) {
	  return function () {
	    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
	      args[_key] = arguments[_key];
	    }
	    return fns.map(function (fn) {
	      return fn.apply(void 0, args);
	    });
	  };
	}

	/**
	 * returns the specified amount of the provided array elements starting from the end
	 * @example
	 * takeLast(2, ['a', 'b', 'c', 'd'])
	 * // returns ['c', 'd']
	 */
	function takeLast(count, arr) {
	  return arr.slice(-count);
	}

	/**
	 * returns a function that called invokes the provided arguments transformers
	 * @example
	 * const func = overArgs((a, b, c) => [a, b, c], [n => n * n, Boolean])
	 * func(2, 0, 1)
	 * // returns [4, false, 1]
	 */
	function overArgs(fn, transformers) {
	  return function () {
	    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
	      args[_key] = arguments[_key];
	    }
	    var transformed = transformers.map(function (transform, index) {
	      return transform(args[index]);
	    });
	    return args.length > transformed.length ? fn.apply(void 0, transformed.concat(takeLast(args.length - transformed.length, args))) : fn.apply(void 0, transformed);
	  };
	}

	/**
	 * returns a tuple from the two provided values
	 * @example
	 * pair('a', 'b')
	 * // returns ['a', 'b']
	 */
	function pair(a, b) {
	  return [a, b];
	}

	/**
	 * returns a tuple of valid and invalid parts of the object as defined in validation function passed in the first argument
	 * @example
	 * partitionObject(el => el % 2 === 0, { a: 1, b: 2, c: 3 })
	 * // returns [{ b: 2 }, { a: 1, c: 3 }]
	 */
	function partitionObject(predicate, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    acc[predicate(obj[key]) ? 0 : 1][key] = obj[key];
	    return acc;
	  }, [{}, {}]);
	}

	/**
	 * picks specified properties from the object
	 * @example
	 * pick(['b'], { a: 1, b: 2 })
	 * // returns { b: 2 }
	 */
	function pick(keys, obj) {
	  return keys.reduce(function (acc, prop) {
	    acc[prop] = obj[prop];
	    return acc;
	  }, {});
	}

	/**
	 * returns an object same as the provided one but with the fields picked by the predicate function
	 * @example
	 * pickBy(el => el % 2 === 0, { a: 1, b: 2, c: 3 })
	 * // returns { b: 2 }
	 */
	function pickBy(predicate, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    if (predicate(obj[key])) {
	      acc[key] = obj[key];
	    }
	    return acc;
	  }, {});
	}

	/**
	 * returns an object same as the provided one but with the fields picked by the predicate function
	 * @example
	 * pickByIndexed((val, key) => key === 'b', { a: 1, b: 2, c: 3 })
	 * // returns { b: 2 }
	 */
	function pickByIndexed(predicate, obj) {
	  return keys(obj).reduce(function (acc, key) {
	    if (predicate(obj[key], key)) {
	      acc[key] = obj[key];
	    }
	    return acc;
	  }, {});
	}

	/**
	 * picks specified props only if they exist on the provided object
	 * @example
	 * pickOwn(['b'], { a: 1, b: 2 })
	 * // returns { b: 2 }
	 */
	function pickOwn(props, obj) {
	  return props.reduce(function (acc, prop) {
	    if (hasOwn(prop, obj)) {
	      acc[prop] = obj[prop];
	    }
	    return acc;
	  }, {});
	}

	/**
	 * returns a random integer number between the specified min and max values
	 * @example
	 * randomInt(0, 10)
	 * // returns 7
	 */
	function randomInt(min, max) {
	  return min + Math.floor(Math.random() * (max - min + 1));
	}

	/**
	 * returns an array filled with numbers in an ascending order to the provided max value
	 * @example
	 * range(5)
	 * // returns [0, 1, 2, 3, 4, 5]
	 */
	function range(max) {
	  var arr = [];
	  var counter = 0;
	  while (counter <= max) {
	    arr.push(counter++);
	  }
	  return arr;
	}

	/**
	 * returns an array with elements that return false for the provided predicate function
	 * @example
	 * reject(el => el % 2 === 0, [1, 2, 3, 4, 5])
	 * // returns [1, 3, 5]
	 */
	function reject(predicate, arr) {
	  return arr.filter(function (element) {
	    return !predicate(element);
	  });
	}

	/**
	 * returns the provided array without the specified index element
	 * @example
	 * removeAt(2, [1, 2, 3, 4, 5])
	 * // returns [1, 2, 4, 5]
	 */
	function removeAt(index, arr) {
	  var copy = [].concat(arr);
	  copy.splice(index, 1);
	  return copy;
	}

	function _extends() {
	  _extends = Object.assign ? Object.assign.bind() : function (target) {
	    for (var i = 1; i < arguments.length; i++) {
	      var source = arguments[i];
	      for (var key in source) {
	        if (Object.prototype.hasOwnProperty.call(source, key)) {
	          target[key] = source[key];
	        }
	      }
	    }
	    return target;
	  };
	  return _extends.apply(this, arguments);
	}
	function _unsupportedIterableToArray(o, minLen) {
	  if (!o) return;
	  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
	  var n = Object.prototype.toString.call(o).slice(8, -1);
	  if (n === "Object" && o.constructor) n = o.constructor.name;
	  if (n === "Map" || n === "Set") return Array.from(o);
	  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
	}
	function _arrayLikeToArray(arr, len) {
	  if (len == null || len > arr.length) len = arr.length;
	  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];
	  return arr2;
	}
	function _createForOfIteratorHelperLoose(o, allowArrayLike) {
	  var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"];
	  if (it) return (it = it.call(o)).next.bind(it);
	  if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
	    if (it) o = it;
	    var i = 0;
	    return function () {
	      if (i >= o.length) return {
	        done: true
	      };
	      return {
	        done: false,
	        value: o[i++]
	      };
	    };
	  }
	  throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
	}

	/**
	 * sets the property of the provided object specified by the provided path in form of a string or an array of keys
	 * @example
	 * const obj = { a: 1, b: { c: 2 } }
	 * set('b.d', 3, obj)
	 * // returns { a: 1, b: { c: 2, d: 3 } }
	 * set(['b', 'd'], 3, obj)
	 * // returns { a: 1, b: { c: 2, d: 3 } }
	 */
	function set(_keys, val, obj) {
	  var _extends2;
	  var keys = isArray(_keys) ? _keys : _keys.split('.');
	  var index = keys[0];
	  var finalVal = val;
	  if (keys.length > 1) {
	    // eslint-disable-next-line eqeqeq
	    var nextObj = obj != null && hasOwn(index, obj) ? obj[index] : {};
	    finalVal = set(keys.slice(1), val, nextObj);
	  }
	  return _extends({}, obj, (_extends2 = {}, _extends2[index] = finalVal, _extends2));
	}

	// https://github.com/reactjs/react-redux/blob/5d792a283554cff3d2f54fad1be1f79cbcab33fe/src/utils/shallowEqual.js
	function is(first, second) {
	  if (first === second) {
	    return first !== 0 || second !== 0 || 1 / first === 1 / second;
	  }
	  // eslint-disable-next-line no-self-compare
	  return first !== first && second !== second;
	}

	/**
	 * returns true if the provided values are shallow equal
	 * @example
	 * shallowEqual({ a: 1 }, { a: 1 })
	 * // returns true
	 * shallowEqual({ a: { b: 1 } }, { a: { b: 1 } })
	 * // returns false
	 */
	function shallowEqual(objA, objB) {
	  if (is(objA, objB)) return true;
	  if (typeof objA !== 'object' || objA === null || typeof objB !== 'object' || objB === null) {
	    return false;
	  }
	  var keysA = Object.keys(objA);
	  var keysB = Object.keys(objB);
	  if (keysA.length !== keysB.length) return false;
	  for (var index = 0; index < keysA.length; index++) {
	    if (!hasOwn(keysA[index], objB) || !is(objA[keysA[index]], objB[keysA[index]])) {
	      return false;
	    }
	  }
	  return true;
	}

	/**
	 * shortens the provided string by the specified amount
	 * @example
	 * shortenLongText(5, 'Lorem ipsum dolor')
	 * // returns 'Lorem...'
	 */
	function shortenLongText(limit, text) {
	  if (text.length <= limit) {
	    return text;
	  }
	  var words = text.split(' ');
	  var sentence = [];
	  var totalLength = 0;
	  for (var _iterator = _createForOfIteratorHelperLoose(words), _step; !(_step = _iterator()).done;) {
	    var word = _step.value;
	    if (totalLength + word.length > limit) {
	      break;
	    }
	    totalLength += word.length + 1;
	    sentence.push(word);
	  }
	  return sentence.join(' ') + "...";
	}

	/**
	 * returns a sign of the provided number or NaN if value different than number was provided
	 * @example
	 * sign(10)
	 * // returns 1
	 * sign(-8)
	 * // returns -1
	 */
	function sign(value) {
	  if (typeof value !== 'number' || isNaN(value)) {
	    return NaN;
	  }
	  if (Object.is(value, 0)) {
	    return 0;
	  }
	  if (Object.is(value, -0)) {
	    return -0;
	  }
	  return value > 0 ? 1 : -1;
	}

	/**
	 * returns an array with the provided array values shuffled
	 * @example
	 * shuffle([1, 2, 3, 4])
	 * // returns [4, 1, 3, 2]
	 */
	function shuffle(arr) {
	  var copy = arr.slice();
	  var lastIndex = arr.length - 1;
	  for (var i = 0; i < arr.length; i++) {
	    var randomIndex = randomInt(i, lastIndex);
	    var randomValue = copy[randomIndex];
	    copy[randomIndex] = copy[i];
	    copy[i] = randomValue;
	  }
	  return copy;
	}

	/**
	 * returns the diff object from the provided object and slice comparison
	 * @example
	 * sliceDiff({ a: 1, b: 10, g: 3 }, { a: 1, b: 2, c: 3 })
	 * // returns { b: 10, g: 3 }
	 */
	function sliceDiff(slice, obj) {
	  var picked = pickByIndexed(function (value, key) {
	    return value !== obj[key];
	  }, slice);
	  return isEmpty(picked) ? null : picked;
	}

	// TODO: this could be written a lot better

	/**
	 * snake cases the provided string
	 * @example
	 * snakeCase('helloWorld')
	 * // returns 'hello_world'
	 */
	function snakeCase(str) {
	  var snakeCased = str.replace(/[A-Z]|([-_ ]+)/g, function (match) {
	    var code = match.charCodeAt(0);
	    var upperCased = code > 64 && code < 91;
	    return upperCased ? "_" + match.toLowerCase() : '_';
	  });
	  return snakeCased[0] === '_' ? snakeCased.substr(1) : snakeCased;
	}

	/**
	 * returns true if some of the array values is truthy
	 * @example
	 * someAreTruthy([0, 1, 2])
	 * // returns true
	 */
	function someAreTruthy(arr) {
	  return arr.some(identity);
	}

	/**
	 * splits an array or a string at the given index
	 * @example
	 * splitAt(2, [1, 2, 3, 4, 5])
	 * // returns [[1, 2], [3, 4, 5]]
	 * splitAt(2, 'foobar')
	 * // returns ['fo', 'obar']
	 */
	function splitAt(splitPoint, arrOrStr) {
	  return [arrOrStr.slice(0, splitPoint), arrOrStr.slice(splitPoint, arrOrStr.length)];
	}

	/**
	 * returns a tuple with the provided array splited on an element that returned true for the provided function, iterating from the right side
	 * @example
	 * splitRightWhenAccum((el, acc) => [el % 2 === 0, acc], [], [1, 2, 3, 4])
	 * // returns [[1, 2, 3], [4]]
	 */
	function splitRightWhenAccum(fn, acc, arr) {
	  var result = false;
	  for (var index = arr.length; index > 0; index--) {
	    var _fn = fn(arr[index - 1], acc);
	    result = _fn[0];
	    acc = _fn[1];
	    if (result) {
	      return splitAt(index - 1, arr);
	    }
	  }
	  return [[], arr];
	}

	/**
	 * returns a function that called with an array of values calls the provided function with the spreaded arguments from the array arguments
	 * @example
	 * const func = spread((a, b, c) => a + b + c)
	 * func([1, 2, 3])
	 * // returns 6
	 */
	function spread(fn) {
	  return function (args) {
	    return fn.apply(void 0, args);
	  };
	}

	/**
	 * returns a sum of all the provided array number values
	 * @example
	 * sum([1, 2, 3])
	 * // returns 6
	 */
	function sum(numbers) {
	  return numbers.reduce(add, 0);
	}

	/**
	 * returns the specified amount of the provided array elements starting from the beginning
	 * @example
	 * take(2, ['a', 'b', 'c', 'd'])
	 * // returns ['a', 'b']
	 */
	function take(count, arr) {
	  return arr.slice(0, count);
	}

	/**
	 * takes elements while `predicate` returns true starting from the right from the specified index
	 * @example
	 * takeRightWhileFrom(el => el > 2, 3, [1, 2, 3, 4, 5])
	 * // returns [3, 4]
	 */
	function takeRightWhileFrom(predicate, startIndex, arr) {
	  var endIndex = findLastIndexFrom(function (element) {
	    return !predicate(element);
	  }, startIndex, arr);
	  return endIndex === startIndex ? [] : arr.slice(endIndex + 1, startIndex + 1);
	}

	/**
	 * takes elements while `predicate` returns true
	 * @example
	 * takeRightWhile(el => el > 2, [1, 2, 3, 4, 5])
	 * // returns [3, 4]
	 */
	function takeRightWhile(predicate, arr) {
	  return takeRightWhileFrom(predicate, arr.length - 1, arr);
	}

	/**
	 * ensures that a function is executed at a fixed interval, so that it is not called too frequently
	 * @example
	 * const updatePreview = () => { ... }
	 * const throttledUpdatePreview = throttle(updatePreview, 500);
	 * inputField.addEventListener('input', throttledUpdatePreview);
	 */
	function throttle(ms, fn) {
	  var lastCall = Date.now() - 2 * ms;
	  var timerId;
	  var invoke = function invoke() {
	    lastCall = Date.now();
	    fn.apply(void 0, arguments);
	  };
	  var cancel = function cancel() {
	    return clearTimeout(timerId);
	  };
	  var throttled = function throttled() {
	    var now = Date.now();
	    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
	      args[_key] = arguments[_key];
	    }
	    if (now - lastCall >= ms) {
	      invoke.apply(void 0, args);
	      return;
	    }
	    cancel();
	    timerId = setTimeout.apply(void 0, [invoke, lastCall - now + ms].concat(args));
	  };
	  throttled.cancel = cancel;
	  return throttled;
	}

	/**
	 * converts values that are iterable to an array
	 * @example
	 * toArray('hello')
	 * // returns ['h', 'e', 'l', 'l', 'o']
	 */
	function toArray(arrayLike) {
	  return Array.prototype.slice.call(arrayLike);
	}

	/**
	 * returns the provided number with specified amount of decimal digits
	 * @example
	 * toFixedNumber(1.2345, 2)
	 * // returns 1.23
	 */
	function toFixedNumber(num, digits) {
	  return Number(num.toFixed(digits));
	}

	/**
	 * returns an array of derived from the provided object key-value tuples
	 * @example
	 * toPairs({ a: 1, b: 2 })
	 * // returns [['a', 1], ['b', 2]]
	 */
	function toPairs(obj) {
	  return keys(obj).map(function (key) {
	    return [key, obj[key]];
	  });
	}

	/**
	 * ensures that a function is executed at a fixed interval, so that it is not called too frequently
	 * @example
	 * const updatePreview = () => { ... }
	 * const throttledUpdatePreview = trailingThrottle(updatePreview, 500);
	 * inputField.addEventListener('input', throttledUpdatePreview);
	 */
	function trailingThrottle(ms, fn) {
	  var lastCall = Date.now() - 2 * ms;
	  var trailingId;
	  var invoke = function invoke() {
	    lastCall = Date.now();
	    return fn.apply(void 0, arguments);
	  };
	  var cancel = function cancel() {
	    return clearTimeout(trailingId);
	  };
	  var throttled = function throttled() {
	    var now = Date.now();
	    if (now - lastCall >= ms) {
	      lastCall = Date.now();
	    }
	    cancel();
	    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
	      args[_key] = arguments[_key];
	    }
	    trailingId = setTimeout.apply(void 0, [invoke, lastCall - now + ms].concat(args));
	  };
	  throttled.cancel = cancel;
	  return throttled;
	}

	/**
	 * ensures that a function is executed at a fixed interval, so that it is not called too frequently
	 * @example
	 * const updatePreview = () => { ... }
	 * const throttledUpdatePreview = leadingThrottle(updatePreview, 500);
	 * inputField.addEventListener('input', throttledUpdatePreview);
	 */
	function leadingThrottle(ms, fn) {
	  var lastCall = 0;
	  return function () {
	    var now = Date.now();
	    if (now - lastCall >= ms) {
	      lastCall = Date.now();
	      fn.apply(void 0, arguments);
	    }
	  };
	}

	/**
	 * returns 0 for matching strings
	 * return -1 for string with lower char code at some position
	 * return 1 for string with greater char code at some position
	 * @example
	 * stringCompare('abc', 'abc')
	 * // returns 0
	 * stringCompare('abc', 'abd')
	 * // returns -1
	 */
	function stringCompare(strA, strB) {
	  if (strA === strB) {
	    return 0;
	  }
	  return strA < strB ? -1 : 1;
	}

	var leadingWhiteSpace = /^\s+/;

	/**
	 * trims the beginning whitespaces from the provided string
	 * @example
	 * trimStart('      hello')
	 * // returns 'hello'
	 */
	function trimStart(str) {
	  return str.replace(leadingWhiteSpace, '');
	}

	var trailingWhiteSpace = /\s+$/;

	/**
	 * trims the end whitespaces from the provided string
	 * @example
	 * trimEnd('hello     ')
	 * // returns 'hello'
	 */
	function trimEnd(str) {
	  return str.replace(trailingWhiteSpace, '');
	}

	/**
	 * returns the provided string repeated the specified amount of times
	 * @example
	 * repeat(3, 'test')
	 * // returns 'testtesttest'
	 */
	function repeat(count, text) {
	  return Array(count + 1).join(text);
	}

	/**
	 * returns an array with all the duplicates from the provided array removed based on the iteratee function
	 * @example
	 * uniqBy(el => el.toString(), [1, '1', 2, '3', 3])
	 * // returns [1, 2, '3']
	 */
	function uniqBy(iteratee, arr) {
	  // with polyfills this could be just: return Array.from(new Set(arr.map(iteratee)))
	  var seen = [];
	  return arr.filter(function (element) {
	    var key = iteratee(element);
	    if (seen.indexOf(key) === -1) {
	      seen.push(key);
	      return true;
	    }
	    return false;
	  });
	}

	/**
	 * returns an array with all the duplicates from the provided array removed
	 * @example
	 * uniq([1, 1, 2, 3, 3])
	 * // returns [1, 2, 3]
	 */
	function uniq(arr) {
	  return uniqBy(identity, arr);
	}

	/**
	 * updates the provided array with the provided value on the specified index
	 * @example
	 * update(2, 3, [1, 2, 5])
	 * // returns [1, 2, 3]
	 */
	function update(index, newElement, arr) {
	  return [].concat(arr.slice(0, index), [newElement], arr.slice(index + 1, arr.length));
	}

	/**
	 * returns an array without the specified elements from the provided array
	 * @example
	 * without([2, 4], [1, 2, 3, 4, 5])
	 * // returns [1, 3, 5]
	 */
	function without(removed, arr) {
	  return arr.filter(function (element) {
	    return removed.indexOf(element) === -1;
	  });
	}

	/**
	 * returns an array of tuples of certain index elements from two provided arrays based on the zipper function
	 * @example
	 * zipWith((a, b) => [a * 2, b.toString()], [1, 2, 3], [10, 20, 30])
	 * // returns [[2, '10'], [4, '20'], [6, '30']]
	 */
	function zipWith(zipper, arrayA, arrayB) {
	  return arrayA.map(function (elementA, index) {
	    return zipper(elementA, arrayB[index]);
	  });
	}

	/**
	 * returns an array of tuples of certain index elements from two provided arrays
	 * @example
	 * zip([1, 2, 3], [10, 20, 30])
	 * // returns [[1, 10], [2, 20], [3, 30]]
	 */
	function zip(arrayA, arrayB) {
	  return zipWith(pair, arrayA, arrayB);
	}

	exports.add = add;
	exports.assign = assign;
	exports.camelCase = camelCase;
	exports.capitalizeFirstLetter = capitalizeFirstLetter;
	exports.chain = chain;
	exports.chunk = chunk;
	exports.cloneDeep = cloneDeep;
	exports.compact = compact;
	exports.compose = compose;
	exports.debounce = debounce;
	exports.deepMapKeys = deepMapKeys;
	exports.deepMerge = deepMerge;
	exports.defaultTo = defaultTo;
	exports.diffKeys = diffKeys;
	exports.drop = drop;
	exports.dropRight = dropRight;
	exports.ensureArray = ensureArray;
	exports.entries = entries;
	exports.filledArray = filledArray;
	exports.find = find;
	exports.findIndex = findIndex;
	exports.findKey = findKey;
	exports.findLast = findLast;
	exports.findLastIndex = findLastIndex;
	exports.findLastIndexFrom = findLastIndexFrom;
	exports.flatMap = flatMap;
	exports.flatten = flatten;
	exports.forOwn = forOwn;
	exports.fromPairs = fromPairs;
	exports.generateRandomId = generateRandomId;
	exports.generateUniqueId = generateUniqueId;
	exports.get = get;
	exports.getOr = getOr;
	exports.groupBy = groupBy;
	exports.groupKeys = groupKeys;
	exports.hasOwn = hasOwn;
	exports.identity = identity;
	exports.includes = includes;
	exports.isArray = isArray;
	exports.isEmpty = isEmpty;
	exports.isFalsy = isFalsy;
	exports.isNil = isNil;
	exports.isObject = isObject;
	exports.isPromise = isPromise;
	exports.isTruthy = isTruthy;
	exports.keyBy = keyBy;
	exports.keys = keys;
	exports.last = last;
	exports.leadingThrottle = leadingThrottle;
	exports.mapKeys = mapKeys;
	exports.mapValues = mapValues;
	exports.mapValuesIndexed = mapValuesIndexed;
	exports.memoize = memoize;
	exports.memoizeOne = memoizeOne;
	exports.memoizeWith = memoizeWith;
	exports.merge = merge;
	exports.mergeAll = mergeAll;
	exports.noop = noop;
	exports.numericSortBy = numericSortBy;
	exports.omit = omit;
	exports.omitBy = omitBy;
	exports.omitByIndexed = omitByIndexed;
	exports.once = once;
	exports.over = over;
	exports.overArgs = overArgs;
	exports.pair = pair;
	exports.partitionObject = partitionObject;
	exports.pick = pick;
	exports.pickBy = pickBy;
	exports.pickByIndexed = pickByIndexed;
	exports.pickOwn = pickOwn;
	exports.randomInt = randomInt;
	exports.range = range;
	exports.reject = reject;
	exports.removeAt = removeAt;
	exports.repeat = repeat;
	exports.set = set;
	exports.shallowEqual = shallowEqual;
	exports.shortenLongText = shortenLongText;
	exports.shuffle = shuffle;
	exports.sign = sign;
	exports.sliceDiff = sliceDiff;
	exports.snakeCase = snakeCase;
	exports.someAreTruthy = someAreTruthy;
	exports.splitAt = splitAt;
	exports.splitRightWhenAccum = splitRightWhenAccum;
	exports.spread = spread;
	exports.stringCompare = stringCompare;
	exports.sum = sum;
	exports.take = take;
	exports.takeLast = takeLast;
	exports.takeRightWhile = takeRightWhile;
	exports.takeRightWhileFrom = takeRightWhileFrom;
	exports.throttle = throttle;
	exports.toArray = toArray;
	exports.toFixedNumber = toFixedNumber;
	exports.toPairs = toPairs;
	exports.trailingThrottle = trailingThrottle;
	exports.trimEnd = trimEnd;
	exports.trimStart = trimStart;
	exports.uniq = uniq;
	exports.uniqBy = uniqBy;
	exports.update = update;
	exports.values = values;
	exports.without = without;
	exports.zip = zip;
	exports.zipWith = zipWith;

	Object.defineProperty(exports, '__esModule', { value: true });

})));
